import os, json

from shapely import buffer, LineString, Point, Polygon, MultiPoint, concave_hull
from typing import List, Tuple
from math import pi

from structs.landmarks import Landmark
from landmarks_manager import take_most_important
from optimizer_v4 import solve_optimization, link_list_simple, print_res, get_time


# Create corridor from tour
def create_corridor(landmarks: List[Landmark], width: float) :

  corrected_width = (180*width)/(6371000*pi)
  
  path = create_linestring(landmarks)
  obj = buffer(path, corrected_width, join_style="mitre", cap_style="square", mitre_limit=2)

  return obj


# Create linestring from tour
def create_linestring(landmarks: List[Landmark])->List[Point] :

  points = []

  for landmark in landmarks :
    points.append(Point(landmark.location))

  return LineString(points)


# Check if some coordinates are in area. Used for the corridor 
def is_in_area(area: Polygon, coordinates) -> bool :
  point = Point(coordinates)
  return point.within(area)


# Function to determine if two landmarks are close to each other 
def is_close_to(location1: Tuple[float], location2: Tuple[float]):
    """Determine if two locations are close by rounding their coordinates to 3 decimals."""
    absx = abs(location1[0] - location2[0])
    absy = abs(location1[1] - location2[1])

    return absx < 0.001 and absy < 0.001
    #return (round(location1[0], 3), round(location1[1], 3)) == (round(location2[0], 3), round(location2[1], 3))


# Rearrange some landmarks in the order of visit to group visit 
def rearrange(landmarks: List[Landmark]) -> List[Landmark]:
  
  i = 1
  while i < len(landmarks):
    j = i+1
    while j < len(landmarks):
      if is_close_to(landmarks[i].location, landmarks[j].location) and landmarks[i].name not in ['start', 'finish'] and landmarks[j].name not in ['start', 'finish']:
        # If they are not adjacent, move the j-th element to be adjacent to the i-th element
        if j != i + 1:
          landmarks.insert(i + 1, landmarks.pop(j))
        break  # Move to the next i-th element after rearrangement
      j += 1
    i += 1
  
  return landmarks


# Simple nearest neighbour planner to try to fix the path
def find_shortest_path_through_all_landmarks(landmarks: List[Landmark]) -> Tuple[List[Landmark], Polygon]:
    
  # Read from data
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']
    
  # Step 1: Find 'start' and 'finish' landmarks
  start_idx = next(i for i, lm in enumerate(landmarks) if lm.name == 'start')
  finish_idx = next(i for i, lm in enumerate(landmarks) if lm.name == 'finish')
  
  start_landmark = landmarks[start_idx]
  finish_landmark = landmarks[finish_idx]
  

  # Step 2: Create a list of unvisited landmarks excluding 'start' and 'finish'
  unvisited_landmarks = [lm for i, lm in enumerate(landmarks) if i not in [start_idx, finish_idx]]
  
  # Step 3: Initialize the path with the 'start' landmark
  path = [start_landmark]
  coordinates = [landmarks[start_idx].location]

  current_landmark = start_landmark
  
  # Step 4: Use nearest neighbor heuristic to visit all landmarks
  while unvisited_landmarks:
    nearest_landmark = min(unvisited_landmarks, key=lambda lm: get_time(current_landmark.location, lm.location, detour, speed))
    path.append(nearest_landmark)
    coordinates.append(nearest_landmark.location)
    current_landmark = nearest_landmark
    unvisited_landmarks.remove(nearest_landmark)

  # Step 5: Finally add the 'finish' landmark to the path
  path.append(finish_landmark)
  coordinates.append(landmarks[finish_idx].location)

  path_poly = Polygon(coordinates)

  return path, path_poly


# Returns a list of minor landmarks around the planned path to enhance experience
def get_minor_landmarks(all_landmarks: List[Landmark], visited_landmarks: List[Landmark], width: float) -> List[Landmark] :

  second_order_landmarks = []
  visited_names = []
  area = create_corridor(visited_landmarks, width)

  for visited in visited_landmarks :
    visited_names.append(visited.name)
  
  for landmark in all_landmarks :
    if is_in_area(area, landmark.location) and landmark.name not in visited_names:
      second_order_landmarks.append(landmark)

  return take_most_important(second_order_landmarks, len(visited_landmarks))


# Try fix the shortest path using shapely
def fix_using_polygon(tour: List[Landmark])-> List[Landmark] :

  coords = []
  coords_dict = {}
  for landmark in tour :
    coords.append(landmark.location)
    if landmark.name != 'finish' :
      coords_dict[landmark.location] = landmark

  tour_poly = Polygon(coords)
  
  better_tour_poly = tour_poly.buffer(0)
  try :
    xs, ys = better_tour_poly.exterior.xy

    if len(xs) != len(tour) :
      better_tour_poly = concave_hull(MultiPoint(coords))  # Create concave hull with "core" of tour leaving out start and finish
      xs, ys = better_tour_poly.exterior.xy

  except :
    better_tour_poly = concave_hull(MultiPoint(coords))  # Create concave hull with "core" of tour leaving out start and finish
    xs, ys = better_tour_poly.exterior.xy


  # reverse the xs and ys
  xs.reverse()
  ys.reverse()

  better_tour = []   # List of ordered visit
  name_index = {}     # Maps the name of a landmark to its index in the concave polygon

  # Loop through the polygon and generate the better (ordered) tour
  for i,x in enumerate(xs[:-1]) :
    y = ys[i]
    better_tour.append(coords_dict[tuple((x,y))])
    name_index[coords_dict[tuple((x,y))].name] = i


  # Scroll the list to have start in front again
  start_index = name_index['start']
  better_tour = better_tour[start_index:] + better_tour[:start_index]

  # Append the finish back and correct the time to reach
  better_tour.append(tour[-1])

  # Rearrange only if polygon still not simple
  if not better_tour_poly.is_simple :
    better_tour = rearrange(better_tour)
  
  return better_tour


# Second stage of the optimization. Use linear programming again to refine the path
def refine_optimization(landmarks: List[Landmark], base_tour: List[Landmark], max_time: int, detour: int, print_infos: bool) -> List[Landmark] :

  # Read from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    max_landmarks = parameters['max landmarks'] + 4

  # No need to refine if no detour is taken
  # if detour == 0 :
  if False :
    new_tour = base_tour
  
  else :
    minor_landmarks = get_minor_landmarks(landmarks, base_tour, 200)

    if print_infos : print("Using " + str(len(minor_landmarks)) + " minor landmarks around the predicted path")

    # full set of visitable landmarks
    full_set = base_tour[:-1] + minor_landmarks   # create full set of possible landmarks (without finish)
    full_set.append(base_tour[-1])                # add finish back

    # get a new tour
    new_tour = solve_optimization(full_set, max_time+detour, False, max_landmarks)
    if new_tour is None :
      new_tour = base_tour

  # Link the new tour
  new_tour, new_dist = link_list_simple(new_tour)

  # If the tour contains only one landmark, return
  if len(new_tour) < 4 :
    return new_tour
  
  # Find shortest path using the nearest neighbor heuristic
  better_tour, better_poly = find_shortest_path_through_all_landmarks(new_tour)

  # Fix the tour using Polygons if the path looks weird
  if base_tour[0].location == base_tour[-1].location and not better_poly.is_valid :
    better_tour = fix_using_polygon(better_tour)
  
  # Link the tour again
  better_tour, better_dist = link_list_simple(better_tour)

  # Choose the better tour depending on walked distance
  if new_dist < better_dist :
    final_tour = new_tour
  else :
    final_tour = better_tour

  if print_infos :
    print("\n\n\nRefined tour (result of second stage optimization): ")
    print_res(final_tour)
    total_score = 0
    for elem in final_tour : 
      total_score += elem.attractiveness

    print("\nTotal score : " + str(total_score))


  
  return final_tour


